/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ASSEMBLYJOBPERFECT_CPP
#define ASSEMBLYJOBPERFECT_CPP


#include "AssemblyJobPerfect.h"

#include "AssemblyJob.h"
#include "AssemblyJobHelperStrategies.h"
#include "AssemblyException.h"
#include <queue>
#include <string>
#include <iostream>
using namespace::std;

AssemblyJobPerfect::AssemblyJobPerfect(){}
AssemblyJobPerfect::AssemblyJobPerfect(set<ScoredSeq*>* seqs,
				       AssemblyJob::AssemblyStrandedness strandedness) :
  _wasRun(false),
  _inputRemoved(false),
  _novelRemoved(false),
  _strandedness(strandedness){
  _inputSeqs.insert(seqs->begin(),seqs->end());
}


AssemblyJobPerfect::~AssemblyJobPerfect(){
}



bool AssemblyJobPerfect::wasRun(){ return _wasRun; }

void AssemblyJobPerfect::allInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
  outsideSet->insert( _usedSeqs.begin(), _usedSeqs.end() );
}

void AssemblyJobPerfect::remainingInputSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobIterative::remainingInputSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobIterative::remainingInputSeqs; seqs were already removed.");
  }
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
}

void AssemblyJobPerfect::discardedInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _usedSeqs.begin(), _usedSeqs.end() );
}
void AssemblyJobPerfect::newlyAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobIterative::newlyAssembledSeqs; job hasn't run yet.");
  }
  if ( _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobIterative::newlyAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}

void AssemblyJobPerfect::allAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobIterative::allAssembledSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved or _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobIterative::allAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}





void AssemblyJobPerfect::OK(){
  if (! _wasRun ){
    if ( _novelSeqs.size() != 0 or _allAssembledSeqs.size() != 0 or _remainingSeqs.size() != 0 ){
      throw AssemblyException::LogicError("AssemblyJobPerfect::OK; there should be no product seqs if job hasn't run.");
    }
    if ( _novelRemoved or _inputRemoved ){
      throw AssemblyException::LogicError("AssemblyJobPerfect::OK; seqs can't have been removed if job hasn't run.");
    }
  }
}







void AssemblyJobPerfect::makeShallow(bool shallowDelete){
  if (_wasRun){
    throw AssemblyException::CallingError("AJNull::makeShallow can't be called after the job was run.");
  }
  set<ScoredSeq*> shallowCopies;
  // buffer all input, then copy
  for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
    (*it)->bottomBuffer();
  }
  for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
    shallowCopies.insert( (*it)->shallowCopy() );
    (*it)->deepUnbuffer();
  }
  if ( shallowDelete){
    for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
      delete (*it);
    }
  }
  // replace the input contents with the shallow copies
  _inputSeqs.clear();
  _inputSeqs.insert( shallowCopies.begin(), shallowCopies.end() );
}

void AssemblyJobPerfect::runJob(AssemblyJob::AssemblyThreadedness threadedness){
  if (! _wasRun){

    map<string,ScoredSeq*> stringToSeq; // the condensed set, indexed by seqeuence

    for (set<ScoredSeq*>::iterator inputIt = _inputSeqs.begin(); inputIt != _inputSeqs.end(); ++inputIt){
      bool senseMatch = false;
      bool antisMatch = false;

      // try for a sense match
      string keySeq = (*inputIt)->getSeq('+');
      map<string,ScoredSeq*>::iterator matchIt = stringToSeq.find(keySeq);
      if ( matchIt!=stringToSeq.end() ){
	senseMatch = true;
	ScoredSeq* newContig = subAssemblyHelper( (*inputIt), matchIt->second , '+' );
	matchIt->second = newContig;
      }

      // if appropriate, try for an antisense match
      if (_strandedness == AssemblyJob::doubleStranded and (! senseMatch) ){
	string revKeySeq = (*inputIt)->getSeq('-');
	matchIt = stringToSeq.find(revKeySeq);
	if ( matchIt!=stringToSeq.end() ){
	  antisMatch = true;
	  ScoredSeq* newContig = subAssemblyHelper( (*inputIt), matchIt->second , '-' );
	  matchIt->second = newContig;
	}
      }

      // no match was found, so just add the seq
      if ( (! senseMatch) and (! antisMatch) ){
	stringToSeq.insert( pair<string,ScoredSeq*> (keySeq, (*inputIt) ) );
      }
    }

    // novel seqs now contains all of those, and used contains all of the seqs that were
    // consumed.  those should be deleted from _inputSeqs
    for (set<ScoredSeq*>::iterator it = _usedSeqs.begin(); it != _usedSeqs.end(); ++it){
      _inputSeqs.erase( (*it) );
    }

    _wasRun = true;
  }
  //OK();
}





// returned seq will be on the same strand as matchSeq
ScoredSeq* AssemblyJobPerfect::subAssemblyHelper(ScoredSeq* newInput, ScoredSeq* matchSeq, char sense){

  // create the new (combined) ScoredSeq
  Alignment* al = new Alignment( matchSeq, newInput, sense ); // order is correct here to get the same sense as the matchSeq
  al->addLinkageBlock( newInput, 0, 0, newInput->size() );
  al->lock();
  ScoredSeq* newContig = AssemblyJobHelperStrategies::combine(al);
  delete al;
  _novelSeqs.insert( newContig );

  // retire the input seq and delete found contig if it was novel, or retire it too
  _usedSeqs.insert(newInput);
  set<ScoredSeq*>::iterator oldSeqIt = _novelSeqs.find( matchSeq );
  if (oldSeqIt == _novelSeqs.end()){
    _usedSeqs.insert(matchSeq);
  } else {
    _novelSeqs.erase(oldSeqIt);
    delete matchSeq;
  } 

  return newContig;
}


#endif
